import logging
import os
from importlib import util

_SUPPORTED_PACKAGES = {}

_FILE_DIRECTORY = os.path.abspath(os.path.dirname(__file__))

__all__ = ["debug", "get_version", "package_available"]


def package_available(name: str):
    """Returns if package is available.

    Args:
        name (str): Name of the package.

    Returns:
        bool: Whether module exists in environment.
    """
    global _SUPPORTED_PACKAGES
    if name not in _SUPPORTED_PACKAGES:
        _SUPPORTED_PACKAGES[name] = util.find_spec(name) is not None
    return _SUPPORTED_PACKAGES[name]


def get_version(package_or_name) -> str:
    """Returns package version.

    Args:
        package_or_name (``module`` or ``str``): Module or name of module.
            This package must have the version accessible through ``<module>.__version__``.

    Returns:
        str: The package version.

    Examples:
        >>> get_version("numpy")
        "1.20.0"
    """
    if isinstance(package_or_name, str):
        if not package_available(package_or_name):
            raise ValueError(f"Package {package_or_name} not available")
        spec = util.find_spec(package_or_name)
        package_or_name = util.module_from_spec(spec)
        spec.loader.exec_module(package_or_name)
    version = package_or_name.__version__
    return version


def sitk_available():
    return package_available("SimpleITK")


def cupy_available():
    if "cupy" not in _SUPPORTED_PACKAGES:
        try:
            import cupy  # noqa
        except ImportError:
            _SUPPORTED_PACKAGES["cupy"] = False
    return package_available("cupy")


def sigpy_available():
    return package_available("sigpy")


def torch_available():
    return package_available("torch")


def output_dir() -> str:
    return os.path.abspath(os.path.join(_FILE_DIRECTORY, "../../.voxel"))


def temp_dir() -> str:
    return os.path.join(output_dir(), "temp")


def log_file_path() -> str:
    return os.path.join(output_dir(), "voxel.log")


def debug(value: bool = None) -> bool:
    """Return (and optionally set) debug mode.

    Args:
        value (bool, optional): If specified, sets the debug status.
            If not specified, debug mode is not set, only returned.

    Returns:
        bool: If ``True``, debug mode is active.

    Raises:
        ValueError: If ``value`` is not a supported value.

    Note:
        Changing the debug state changes the stream handler logging level
        for the default voxel logger. If debug state is turned off, logging
        level is set to ``logging.INFO``. If debug state is turned on,
        logging level is set to ``logging.DEBUG``.

    Examples:
        >>> debug()  # get debug status, defaults to False
        False
        >>> debug(True)  # turn on debug mode
        True
        >>> debug()  # get debug status
        True
    """

    def _is_debug():
        return os.environ.get("VOXEL_DEBUG", "") in ["True", "true"]

    def _toggle_debug(_old_value, _new_value):
        # TODO: Toggle voxel logging to debug mode.
        if _old_value == _new_value:
            return

        _vx_logger = logging.getLogger("voxel")
        if _new_value:
            _vx_logger.setLevel(logging.DEBUG)
            for h in _vx_logger.handlers:
                h.setLevel(logging.DEBUG)
        else:
            _vx_logger.setLevel(logging.DEBUG)  # the root logger should always log at DEBUG level
            for h in _vx_logger.handlers:
                if isinstance(h, logging.StreamHandler):
                    h.setLevel(logging.INFO)

    if value is not None:
        old_value = _is_debug()
        if isinstance(value, bool):
            os.environ["VOXEL_DEBUG"] = str(value)
        elif isinstance(value, str) and value.lower() in ("true", "false", ""):
            os.environ["VOXEL_DEBUG"] = value
        else:
            raise ValueError(f"Unknown value for debug: '{value}'")

        _toggle_debug(old_value, _is_debug())

    return _is_debug()



"""Voxel."""
from voxel.device import Device, cpu_device, get_array_module, get_device, to_device  # noqa: F401
from voxel.io import load, read, save, write  # noqa: F401
from voxel.io.dicom import DicomReader, DicomWriter  # noqa: F401
from voxel.io.format_io import ImageDataFormat  # noqa: F401
from voxel.io.http import HttpReader  # noqa: F401
from voxel.io.nifti import NiftiReader, NiftiWriter  # noqa: F401
from voxel.med_volume import MedicalVolume  # noqa: F401
from voxel.orientation import to_affine  # noqa: F401
from voxel.utils.collect_env import collect_env_info  # noqa

from .config import config  # noqa: F401

__all__ = [
    "MedicalVolume",
    "DicomReader",
    "DicomWriter",
    "HttpReader",
    "NiftiReader",
    "NiftiWriter",
    "read",
    "save",
    "load",
    "write",
    "ImageDataFormat",
    "Device",
    "get_device",
    "to_device",
    "cpu_device",
    "get_array_module",
    "to_affine",
    "collect_env_info",
]

# This line will be programatically read/write by setup.py.
# Leave them at the bottom of this file and don't touch them.
__version__ = "0.0.2"



import numpy as np

from voxel.device import cpu_device
from voxel.med_volume import MedicalVolume

try:
    from voxelviz.widget import VoxelWidget

    _SUPPORTS_VOXELVIZ = True
except ImportError:
    widget = None
    _SUPPORTS_VOXELVIZ = False

# `True` if the user is running code in an iPython session.
# This is required to be able to use the viewer.
_IS_IPYTHON_SESSION = None


class Axis:
    def __init__(self) -> None:
        if not _SUPPORTS_VOXELVIZ:
            # TODO: I think the user also has to install ipywidgets.
            # TODO: Add figsize
            raise ImportError(
                "`voxelviz` is not installed. To install, use `pip install voxelviz`."
            )
        if not _is_ipython_session():
            raise RuntimeError("The viewer can only be used in an iPython session.")

    def show(self, volume: MedicalVolume, *, seg: MedicalVolume = None):
        """Display a multi-dimensional volume with optional segmentation overlay.

        Args:
            volume (MedicalVolume): The volume to display.
            seg (MedicalVolume, optional): The segmentation to overlay.
                If specified, all entries must be non-negative integers.
                ``0`` corresponds to background.
        """
        if any(x is not None and x.device != cpu_device for x in [volume, seg]):
            raise ValueError("All images to display must be on the CPU.")

        if volume.ndim > 4:
            raise ValueError("Only volumes with 4 or fewer dimensions are supported.")

        if seg is not None:
            seg = seg.reformat_as(volume)
            # Duplicate the segmentation along the extra dimension.
            # TODO: Does the extra dimension must be the same between the base volume
            # and segmentation?
            if volume.ndim > seg.ndim:
                seg = np.stack([seg] * volume.shape[-1], axis=-1)
            if not seg.is_same_dimensions(volume):
                raise ValueError(
                    "The segmentation and volume must have the same dimensions and spacing."
                )
            if np.any(seg.A < 0):
                raise ValueError("The segmentation must be non-negative.")

        spacing = volume.pixel_spacing
        volume = _auto_reshape(volume)

        vw = VoxelWidget()
        vw.vol = volume
        if seg is not None:
            seg = _auto_reshape(seg)
            vw.seg = seg

        vw.cfg = {"spacing": list(spacing)}
        return vw


def show(volume: MedicalVolume, *, seg: MedicalVolume = None):
    """Display a multi-dimensional volume with optional segmentation overlay.

    Args:
        volume (MedicalVolume): The volume to display.
        seg (MedicalVolume, optional): The segmentation to overlay.
            If specified, all entries must be non-negative integers.
            ``0`` corresponds to background.
    """
    ax = Axis()
    return ax.show(volume, seg=seg)


def _auto_reshape(mv: MedicalVolume):
    """

    The first 3 dimensions in a MedicalVolume are always spatial dimensions.
    We assume we are only working with grayscale images for now.
    """
    assert mv.ndim in (3, 4)
    arr = mv.A

    # Add batch dimension
    if arr.ndim == 3:
        arr = arr[np.newaxis, ...]
    else:
        arr = arr.transpose(3, 0, 1, 2)

    # Add channel dimension.
    arr = arr[..., np.newaxis]
    return arr


def _is_ipython_session() -> bool:
    """Returns if the python is an iPython session.

    Adapted from
    https://discourse.jupyter.org/t/find-out-if-my-code-runs-inside-a-notebook-or-jupyter-lab/6935/3
    """
    global _IS_IPYTHON_SESSION
    if _IS_IPYTHON_SESSION is not None:
        return _IS_IPYTHON_SESSION

    is_ipython_session = None
    try:
        from IPython import get_ipython

        ip = get_ipython()
        is_ipython_session = ip is not None
    except ImportError:
        # iPython is not installed
        is_ipython_session = False
    _IS_IPYTHON_SESSION = is_ipython_session
    return _IS_IPYTHON_SESSION



#!/usr/bin/env python

import os
import subprocess
import sys
from shutil import rmtree

from setuptools import Command, find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))


def get_version():
    init_py_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "voxel", "__init__.py")
    init_py = open(init_py_path, "r").readlines()
    version_line = [l.strip() for l in init_py if l.startswith("__version__")][0]  # noqa: E741
    version = version_line.split("=")[-1].strip().strip("'\"")

    # The following is used to build release packages.
    # Users should never use it.
    suffix = os.getenv("VOXEL_VERSION_SUFFIX", "")
    version = version + suffix
    if os.getenv("BUILD_NIGHTLY", "0") == "1":
        from datetime import datetime

        date_str = datetime.today().strftime("%y%m%d")
        version = version + ".dev" + date_str

        new_init_py = [l for l in init_py if not l.startswith("__version__")]  # noqa: E741
        new_init_py.append('__version__ = "{}"\n'.format(version))
        with open(init_py_path, "w") as f:
            f.write("".join(new_init_py))
    return version


class UploadCommand(Command):
    """Support setup.py upload.

    Only run upload from the main branch.

    Adapted from https://github.com/robustness-gym/meerkat.
    """

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Only upload from the main branch
        branches = subprocess.getoutput("git branch").split("\n")
        branches = [x.strip() for x in branches]
        curr_branch = [x for x in branches if x.startswith("*")]
        if len(curr_branch) != 1:
            raise RuntimeError("Could not determine current branch.")
        curr_branch = curr_branch[0].split(" ")[-1]
        if curr_branch != "main":
            raise RuntimeError("Uploads only allowed from main branch.")

        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel --universal".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(get_version()))
        os.system("git push --tags")

        sys.exit()


# ---------------------------------------------------
# Setup Information
# ---------------------------------------------------

# Required pacakges.
REQUIRED = [
    "dataclasses>=0.6",
    "numpy",
    "natsort",
    "nibabel",
    "packaging",
    "pydicom>=1.6.0",
    "PyYAML>=5.4.1",
    "requests",
    "tabulate",
    "termcolor",
    "tqdm>=4.42.0",
]

# Optional packages.
# TODO Issue #106: Fix to only import tensorflow version with fixed version
# once keras import statements are properly handled.
EXTRAS = {
    "dev": [
        # optional dependency libraries.
        "simpleitk",
        "sigpy",
        "h5py",
        # formatting.
        "coverage",
        "flake8",
        "flake8-bugbear",
        "flake8-comprehensions",
        "isort",
        "black==22.8.0",
        "click==8.0.2",
        # testing.
        "pytest-cov>=2.10.1",
        "pre-commit>=2.9.3",
        "requests-mock",
        "parameterized",
        # upload.
        "twine",
    ],
    "docs": ["mistune>=0.8.1,<2.0.0", "sphinx", "sphinxcontrib.bibtex", "m2r2"],
}

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="pyvoxel",
    version=get_version(),
    author="Arjun Desai",
    url="https://github.com/pyvoxel/pyvoxel",
    project_urls={"Documentation": "https://pyvoxel.readthedocs.io/"},
    description="An AI-powered open-source medical image analysis toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("configs", "tests", "tests.*")),
    python_requires=">=3.6",
    install_requires=REQUIRED,
    license="GNU",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    extras_require=EXTRAS,
    # $ setup.py publish support.
    cmdclass={
        "upload": UploadCommand,
    },
)



"""I/O formatting templates.

This module consists of the templates for input/output (I/O) helper classes.
"""
import enum
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Collection, Dict, Union

__all__ = ["ImageDataFormat", "DataReader", "DataWriter"]


class ImageDataFormat(enum.Enum):
    """Enum describing supported data formats for medical volume I/O."""

    nifti = 1, ("nii", "nii.gz")
    dicom = 2, ("dcm", "ima", "dicom", "dic")

    def __new__(cls, key_code, extensions):
        """
        Args:
            key_code (int): Enum value.
            extensions (tuple[str]): Extensions supported by format.
        """
        obj = object.__new__(cls)
        obj._value_ = key_code
        obj.extensions = extensions
        return obj

    def is_filetype(self, file_path: Union[str, Path, os.PathLike]) -> bool:
        """Verify if file path matches the file type specified by ImageDataFormat.

        This method checks to make sure the extensions are appropriate.

        Args:
            file_path (str): File path.

        Returns:
            bool: True if file_path has valid extension, False otherwise.
        """
        file_path = str(file_path)
        bool_list = [file_path.lower().endswith(".%s" % ext.lower()) for ext in self.extensions]

        return bool(sum(bool_list))

    @classmethod
    def get_image_data_format(cls, file_or_dir_path: Union[str, Path, os.PathLike]):
        """Get the `ImageDataFormat` that corresponds to the file path.

        Matches extension to file path. If input is a directory path, then
        it is classified as `ImageDataFormat.dicom`.

        Args:
            file_or_dir_path (str): Path to a file or a directory.

        Returns:
            ImageDataFormat: Format corresponding to file or directory path.

        Raises:
            ValueError: If no compatible ImageDataFormat found.
        """
        for im_data_format in cls:
            if im_data_format.is_filetype(file_or_dir_path):
                return im_data_format

        # if no extension found, assume the name corresponds to a directory
        # and assume that format is dicom.
        # We cannot check if the path is a directory path because it may not
        # have been created yet.
        file_or_dir_path = str(file_or_dir_path)
        filename_base, ext = os.path.splitext(file_or_dir_path)
        if filename_base == file_or_dir_path or os.path.isdir(file_or_dir_path):
            return ImageDataFormat.dicom

        raise ValueError(f"Unknown data format for {file_or_dir_path}")


class _StateMixin(ABC):
    """Temporary mixin that supports fetching and loading from state dictionaries.

    Note:
        This functionality is not well supported and will likely undergo changes
        in the alpha development process. Use with caution.
    """

    @abstractmethod
    def __serializable_variables__(self) -> Collection[str]:
        """Collection of serializable variables (i.e. state keys).

        This method should be implemented by subclasses and should
        return what parameters can be serialized. The values should
        typically be of a raw type.

        Returns:
            Collection[str]: Serializable variables.
        """
        raise NotImplementedError  # pragma: no cover

    def state_dict(self) -> Dict[str, Any]:
        """Returns the state dictionary for an object.

        Keys for this dictionary are specified by the object's
        :func:`self.__serializable_variables__()` method.

        Returns:
            Dict[str, Any]: The state dictionary.
        """
        return {k: self.__dict__[k] for k in self.__serializable_variables__()}

    def load_state_dict(self, state_dict: Dict[str, Any]):
        """Loads the state dictionary into the object.

        Args:
            state_dict (Dict[str, Any]): The state dictionary created using
                :func:`self.state_dict()`.

        Raises:
            AttributeError: If any key in ``state_dict`` is not an attribute in
                ``self``.
        """
        for k, v in state_dict.items():
            if not hasattr(self, k):
                raise AttributeError(f"{type(self)} does not have attribute '{k}'")
            setattr(self, k, v)


class DataReader(_StateMixin):
    """Abstract class for reading medical data.

    Format-specific readers should inherit from this class.

    Attributes:
        data_format_code (ImageDataFormat): Should be defined by subclasses.
    """

    data_format_code = None

    @abstractmethod
    def load(self, file_path: str):
        """Load volume.

        Args:
            file_path (str): File path to load volume from.

        Returns:
            MedicalVolume: The loaded volume.
        """
        pass  # pragma: no cover

    read = load  # pragma: no cover

    def __call__(self, *args, **kwargs):
        """Alias for :meth:`self.load`."""
        return self.load(*args, **kwargs)


class DataWriter(_StateMixin):
    """Abstract class for writing medical data.

    Format-specific writers should inherit from this class.

    Attributes:
        data_format_code (ImageDataFormat): Should be defined by subclasses.
    """

    data_format_code = None

    @abstractmethod
    def save(self, volume, file_path: str):
        """Save volume.

        Args:
            volume (MedicalVolume): Volume to save.
            file_path (str): File path to save volume to.
        """
        pass  # pragma: no cover

    write = save  # pragma: no cover

    def __call__(self, *args, **kwargs):
        """Alias for :meth:`self.save`."""
        return self.save(*args, **kwargs)
